MODULE likelihood

  USE GLOBVAR
  USE RANDOM
  
  IMPLICIT NONE

    CONTAINS

  SUBROUTINE initialize(ii)

  IMPLICIT NONE

  INTEGER, INTENT(IN) :: ii
  INTEGER :: i,s,ss,b

  DO i=1,N_en
      s = dta2(i,4)
      ss = dta2(i,5)
      IF(dta2(i,2)<thre(1)) THEN  
          IF(dta2(i,3)==0.0d0) dta0(s,ss) = 1
          IF(dta2(i,3)==1.0d0) dta1(s,ss) = 1
      END IF
      DO b=2,N_bin-1
          IF(dta2(i,2)<thre(b) .AND. dta2(i,2)>=thre(b-1)) THEN
              IF(dta2(i,3)==0.0d0) dta0(s,ss) = b
              IF(dta2(i,3)==1.0d0) dta1(s,ss) = b
          END IF
      END DO
      IF(dta2(i,2)>=thre(N_bin-1)) THEN
          IF(dta2(i,3)==0.0d0) dta0(s,ss) = N_bin
          IF(dta2(i,3)==1.0d0) dta1(s,ss) = N_bin
      END IF
  END DO
  initial(:,0) = dta0(:,1)
  initial(:,1) = dta1(:,1)
  
  END SUBROUTINE initialize
    
  
  SUBROUTINE frequency(ii)
  ! Compute q_pool,q_obs,p_pool,F_j_ss from dta2

  IMPLICIT NONE

  INTEGER, INTENT(IN) :: ii
  REAL(8) :: temp,PP(N_bin,N_bin)
  INTEGER :: bin(N_bin),d,b,i,m,n,bb1,bb2,j,s,ss,t

  q_pool=0.0d0
  q_obs=0.0d0
  p_pool=0.0d0
  F_j_ss=0.0d0
  DO m=1,N_market
      DO t=1,T0-1
          F_j_ss(m,dta0(m,t),dta0(m,t+1),0) = F_j_ss(m,dta0(m,t),dta0(m,t+1),0) + 1.0d0          
      END DO
      DO t=1,T1-1
          F_j_ss(m,dta1(m,t),dta1(m,t+1),1) = F_j_ss(m,dta1(m,t),dta1(m,t+1),1) + 1.0d0
      END DO
      DO bb1=1,N_bin
          q_obs(m,bb1,0) = DBLE(count(dta0(m,:)==bb1))/DBLE(T0)
      END DO
      DO bb1=1,N_bin
          q_obs(m,bb1,1) = DBLE(count(dta1(m,:)==bb1))/DBLE(T1)
      END DO
      q_pool(:,0) = q_pool(:,0) + q_obs(m,:,0)        
      q_pool(:,1) = q_pool(:,1) + q_obs(m,:,1)              
  END DO
  q_pool=q_pool/DBLE(N_market)
  DO d=0,1
      DO bb1=1,N_bin
          DO bb2=1,N_bin
              p_pool(bb1,bb2,d) = SUM(F_j_ss(:,bb1,bb2,d))
          END DO
          IF(SUM(p_pool(bb1,:,d))>0.0d0) p_pool(bb1,:,d) = p_pool(bb1,:,d)/SUM(p_pool(bb1,:,d))
      END DO
  END DO
  
  ! Calculate Qsst and iotaP for Ts(s')  
  Qsst = 0.0d0
  DO bb1=1,N_bin
      DO bb2=1,N_bin
          DO t=2,T0
              s = COUNT(dta0(:,1)==bb1 .AND. dta0(:,t)==bb2)
              ss =COUNT(dta0(:,1)==bb1)
              IF(ss>0) Qsst(bb1,bb2,t,0) = DBLE(s)/DBLE(ss)
          END DO
          DO t=2,T1
              s = COUNT(dta1(:,1)==bb1 .AND. dta1(:,t)==bb2)
              ss =COUNT(dta1(:,1)==bb1)
              IF(ss>0) Qsst(bb1,bb2,t,1) = DBLE(s)/DBLE(ss)
          END DO          
      END DO
  END DO  
  iotaP = 0.0d0
  DO t=1,T0
      PP=P_pool(:,:,0)
      DO s=2,t
          PP = MATMUL(PP,P_pool(:,:,0))
      END DO
      iotaP(:,:,t,0) = PP
  END DO
  DO t=1,T1
      PP=P_pool(:,:,1)
      DO s=2,t
          PP = MATMUL(PP,P_pool(:,:,1))
      END DO
      iotaP(:,:,t,1) = PP
  END DO  
  
  END SUBROUTINE frequency


  FUNCTION chi_squared(ii)

  ! Use q_pool, q_obs, p_pool, and F_j_ss to compute the test statistics
  
  IMPLICIT NONE

  INTEGER, INTENT(IN) :: ii
  INTEGER :: bin(N_bin),d,b,s,i,m,n,s1,s2
  REAL(8) :: chi_squared(N_stat)
  REAL(8) :: temp1,temp,temp2,F_j_s(N_market,N_bin,0:1)

  DO d=0,1
      temp1=0.0d0
      DO m=1,N_market
          DO s=1,N_bin
              temp = q_pool(s,d) - q_obs(m,s,d)
              temp1 = temp1 + temp*temp                    
          END DO
      END DO
      IF(d==0) chi_squared(1) = temp1*DBLE(T0)      
      IF(d==1) chi_squared(2) = temp1*DBLE(T1)      
  END DO

  DO d=0,1
      temp1=0.0d0
      DO m=1,N_market
          DO s1=1,N_bin
              temp2 = SUM(F_j_ss(m,s1,:,d))       ! F^j_(s)
              DO s2=1,N_bin
                  temp = F_j_ss(m,s1,s2,d) - temp2*p_pool(s1,s2,d)
                  IF(temp2*p_pool(s1,s2,d)>0.0d0) temp1 = temp1 + temp*temp/(temp2*p_pool(s1,s2,d))
              END DO
          END DO
      END DO
      chi_squared(d+3) = temp1
  END DO
  
  DO d=0,1
      DO m=1,N_market
          DO s1=1,N_bin
              F_j_s(m,s1,d) = SUM(F_j_ss(m,s1,:,d))  
          END DO
      END DO
  END DO
  
  DO d=0,1
      DO m=1,N_market
          DO s1=1,N_bin
              IF(SUM(F_j_ss(m,s1,:,d))>0.0d0) THEN 
                  F_j_ss(m,s1,:,d) = F_j_ss(m,s1,:,d)/SUM(F_j_ss(m,s1,:,d))
              ELSE
                  F_j_ss(m,s1,:,d) = 0.0d0
              END IF
          END DO
      END DO
  END DO
  
  DO d=0,1
      temp1=0.0d0
      DO m=1,N_market
          DO s1=1,N_bin
              DO s2=1,N_bin
                  IF(F_j_ss(m,s1,s2,d)>0.0d0 .AND. p_pool(s1,s2,d)>0.0d0) THEN
                      temp = F_j_ss(m,s1,s2,d)/p_pool(s1,s2,d)
                      temp1 = temp1 + F_j_s(m,s1,d)*F_j_ss(m,s1,s2,d)*LOG(temp)
                  END IF
              END DO
          END DO
      END DO
      chi_squared(d+5) = temp1*2.0d0
  END DO  
  
  DO d=0,1
      temp1=0.0d0
      DO s1=1,N_bin
          DO s2=2,N_bin
              IF(d==0) temp = (SUM(Qsst(s1,s2,:,d)) - SUM(iotaP(s1,s2,:,d)))/DBLE(T0)
              IF(d==1) temp = (SUM(Qsst(s1,s2,:,d)) - SUM(iotaP(s1,s2,:,d)))/DBLE(T1)
              temp1 = temp1 + temp*temp
          END DO
      END DO
      chi_squared(d+7) = temp1*DBLE(N_market)
  END DO
    
  END FUNCTION chi_squared    

  
  SUBROUTINE simulate_game(ii)

  ! Use q_pool and p_pool, then calcualte new (q_pool,p_pool,q_obs,F_j_ss)
  
  IMPLICIT NONE

  INTEGER, INTENT(IN) :: ii
  INTEGER :: s,n,t,i,d,m,j,b,tt
  REAL(8) :: temp,u

  DO m=1,N_market
      DO d=0,1
          u = Sample_Uniform(0.0d0,1.0d0)
          temp=0.0d0    
          DO b=1,N_bin
              temp = temp + q_pool(b,d)
              IF(temp>u) THEN
                  i = b
                  EXIT
              END IF
          END DO
          i = initial(m,d)          
          tt=T0
          IF(d==1) tt=T1
          DO t=1,tt
              IF(d==0) dta0(m,t) = i
              IF(d==1) dta1(m,t) = i
              u = Sample_Uniform(0.0d0,1.0d0)
              temp = 0.0d0
              DO b=1,N_bin
                  temp = temp + p_pool(i,b,d)
                  IF(temp>u) THEN
                      i = b
                      EXIT
                  END IF
              END DO
          END DO
      END DO
  END DO    

  END SUBROUTINE simulate_game

  
END MODULE likelihood

